#pragma once
#include<winhttp.h>
#include<winerror.h>
#include<iostream>
#include<vector>
#include <stdio.h>
#include <curl/curl.h>
#include <nlohmann/json.hpp>
#define CHIMEGE_API_h

using json = nlohmann::json;

static size_t StrWriteCallback(char* contents, size_t size, size_t nmemb, void* userp)
{
	std::string* str = (std::string*)userp;
	std::string chunk = "";
	chunk.assign(contents, contents + size * nmemb);
	*str += chunk;
	return size * nmemb;
}

namespace chimege {
	class ChimegeAPI
	{
	public:
		ChimegeAPI::ChimegeAPI() {}
		~ChimegeAPI() {}
		
		/// <summary>
		///  Normalize and split text for synthesize.
		/// </summary>
		/// <param name="text">Text to be normalized</param>
		/// <param name="result">Result pointer</param>
		/// <returns>HRESULT</returns>
		HRESULT static ChimegeAPI::normalize(std::string text, std::vector<std::string>* result) {
			HRESULT hr = S_OK;

			json jsonData = {
			  {"data", text},
			  {"options", {
				{"split_sentences", true },
				{"abbreviation_level", "abbreviation"},
				{"read_symbols", "no_temdegt"},
				{"read_emojis", true },
				{"use_phonemizer", true },
			  }}
			};

			CURL* curl;
			CURLcode res;

			/* In windows, this will init the winsock stuff */
			curl_global_init(CURL_GLOBAL_ALL);

			/* get a curl handle */
			curl = curl_easy_init();
			if (curl) {
				std::string resJSON = "";
				struct curl_slist* headers = NULL;
				headers = curl_slist_append(headers, "Content-Type: application/json");
				headers = curl_slist_append(headers, "Key: 9XZiFiU7z05pba6yMz8bnlTTQN1IjOx6GxMFHzU");

				std::string jsonStr = jsonData.dump();
				const char* data = jsonStr.c_str();

				curl_easy_setopt(curl, CURLOPT_URL, "http://localhost:6969/tts_normalize");
				curl_easy_setopt(curl, CURLOPT_POST, 1);
				curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data);
				curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, strlen(data));
				curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, StrWriteCallback);
				curl_easy_setopt(curl, CURLOPT_WRITEDATA, &resJSON);
				curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);

				/* Perform the request, res will get the return code */
				res = curl_easy_perform(curl);
				/* Check for errors */
				if (res != CURLE_OK)
					fprintf(stderr, "curl_easy_perform() failed: %s\n",
						curl_easy_strerror(res));

				json res = json::parse(resJSON);
				*result = res["normalized"].get<std::vector<std::string>>();

				/* always cleanup */
				curl_easy_cleanup(curl);
			}
			curl_global_cleanup();

			return hr;
		}

		/// <summary>
		/// Synthesizes audio and returns audio as a buffer.
		/// </summary>
		/// <param name="text">Text to synthesize</param>
		/// <param name="buf">Audio buffer</param>
		/// <returns>HRESULT</returns>
		HRESULT ChimegeAPI::synthesize(std::string text, std::string* buf, double rate, double volume, int pitch, std::string voiceID) {
			HRESULT hr = S_OK;

			json jsonData = {
			  {"voiceID", voiceID}, // Current available voice IDs are 0 to 5
			  {"text", text},
			  {"tempo", rate},
			  {"volume", volume},
			  {"pitch", pitch},
			  {"silence", 0},
			};

			CURL* curl;
			CURLcode res;

			/* In windows, this will init the winsock stuff */
			curl_global_init(CURL_GLOBAL_ALL);

			/* get a curl handle */
			curl = curl_easy_init();
			if (curl) {
				struct curl_slist* headers = NULL;
				headers = curl_slist_append(headers, "Content-Type: application/json");

				std::string jsonStr = jsonData.dump();
				const char* data = jsonStr.c_str();

				curl_easy_setopt(curl, CURLOPT_URL, "http://localhost:6969/synthesize");
				curl_easy_setopt(curl, CURLOPT_POST, 1);
				curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data);
				curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, strlen(data));
				curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, StrWriteCallback);
				curl_easy_setopt(curl, CURLOPT_WRITEDATA, buf);
				curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);

				/* Perform the request, res will get the return code */
				res = curl_easy_perform(curl);
				/* Check for errors */
				if (res != CURLE_OK)
					fprintf(stderr, "curl_easy_perform() failed: %s\n",
						curl_easy_strerror(res));

				/* always cleanup */
				curl_easy_cleanup(curl);
			}
			curl_global_cleanup();

			return hr;
		}
	};
}
