#include "util.h"
#include "stdafx.h"
#include "sapiddk.h"
#include <iostream>
#include <string>
#include <codecvt>
using namespace std;

void GetUnicodeChar(unsigned int code, char chars[5]) {
    if (code <= 0x7F) {
        chars[0] = (code & 0x7F); chars[1] = '\0';
    } else if (code <= 0x7FF) {
        // one continuation byte
        chars[1] = 0x80 | (code & 0x3F); code = (code >> 6);
        chars[0] = 0xC0 | (code & 0x1F); chars[2] = '\0';
    } else if (code <= 0xFFFF) {
        // two continuation bytes
        chars[2] = 0x80 | (code & 0x3F); code = (code >> 6);
        chars[1] = 0x80 | (code & 0x3F); code = (code >> 6);
        chars[0] = 0xE0 | (code & 0xF); chars[3] = '\0';
    } else if (code <= 0x10FFFF) {
        // three continuation bytes
        chars[3] = 0x80 | (code & 0x3F); code = (code >> 6);
        chars[2] = 0x80 | (code & 0x3F); code = (code >> 6);
        chars[1] = 0x80 | (code & 0x3F); code = (code >> 6);
        chars[0] = 0xF0 | (code & 0x7); chars[4] = '\0';
    } else {
        // unicode replacement character
        chars[2] = 0xEF; chars[1] = 0xBF; chars[0] = 0xBD;
        chars[3] = '\0';
    }
}

wstring StringToWString(string str) {
    wstring_convert<codecvt_utf8_utf16<wchar_t>> converter;
    wstring wide = converter.from_bytes(str);
    return wide;
}

string WStringToString(wstring str) {
    wstring_convert<codecvt_utf8_utf16<wchar_t>> converter;
    string narrow = converter.to_bytes(str);
    return narrow;
}

string GetStringFromWinAPI32TextFrag(const SPVTEXTFRAG* pTextFragList) {
    char arr[5];
    string str = "";

    for (int i = 0; i < pTextFragList->ulTextLen; i++) {
        GetUnicodeChar(*(pTextFragList->pTextStart + i), arr);
        str += arr;
    }
    return str;
}

wchar_t const* const digitTables[] =
{
    L"0123456789",
    L"\u0660\u0661\u0662\u0663\u0664\u0665\u0666\u0667\u0668\u0669",
    // ...
};

//!     \return
//!         wch as a numeric digit, or -1 if it is not a digit
int asNumeric(wchar_t wch)
{
    int result = -1;
    for (wchar_t const* const* p = std::begin(digitTables);
        p != std::end(digitTables) && result == -1;
        ++p) {
        wchar_t const* q = std::find(*p, *p + 10, wch);
        if (q != *p + 10) {
            result = q - *p;
        }
    }
    return result;
}